///
/// Extension of the standard class dart [Object]
///
extension ObjectExtension on Object {
  ///
  /// Converts any object to a human-readable structured string.
  ///
  /// You can use the [stringQuote] and [keyQuote] parameters to specify the
  /// quotes character for keys and string values. For ease of debugging, it is
  /// convenient to use [withTypes]. It shows types for all data. [tabWidth]
  /// defines the width of the tabs.
  ///
  /// Lists with simple elements of the same type are shown on one line.
  /// The same is true for sets. Map entries are always shown each on a
  /// separate line.
  ///
  String toStructuredString(
          {String stringQuote = "'",
          String keyQuote = '',
          bool withTypes = false,
          int tabWidth = 2}) =>
      _toStructuredString(
          value: this,
          parentIsSimpleList: false,
          stringQuote: stringQuote,
          keyQuote: keyQuote,
          withTypes: withTypes,
          tabWidth: tabWidth,
          level: 0);

  ///
  /// Recursive variant of [toStructuredString()] method.
  ///
  /// Closed method, used only inside [ObjectExtension]. [value] is the current
  /// nesting element in the object structure. [parentIsSimpleList] is needed to
  /// cancel line wrapping in lists with simple identical objects. [level]
  /// defines the current nesting level in the object structure.
  ///
  /// The [stringQuote], [keyQuote], [withTypes] and [tabWidth] parameters are
  /// described in the reference for the [toStructuredString()] method.
  ///
  String _toStructuredString(
      {required Object value,
      required bool parentIsSimpleList,
      required String stringQuote,
      required String keyQuote,
      required bool withTypes,
      required tabWidth,
      required int level}) {
    final type = withTypes && !parentIsSimpleList
        ? value.runtimeType.toString() + ' '
        : '';

    if (value is Map) {
      return type +
          _mapToStructuredString(
              map: value,
              stringQuote: stringQuote,
              keyQuote: keyQuote,
              withTypes: withTypes,
              tabWidth: tabWidth,
              level: level);
    }
    if (value is List) {
      return type +
          _listToStructuredString(
              list: value,
              stringQuote: stringQuote,
              keyQuote: keyQuote,
              withTypes: withTypes,
              tabWidth: tabWidth,
              level: level);
    }
    if (value is Set) {
      return type +
          _setToStructuredString(
              set: value,
              stringQuote: stringQuote,
              keyQuote: keyQuote,
              withTypes: withTypes,
              tabWidth: tabWidth,
              level: level);
    }

    if (value is String) {
      return type + stringQuote + value + stringQuote;
    }
    return type + value.toString();
  }

  ///
  /// Converts [Map] type object to a human-readable structured string.
  ///
  /// Closed method, used only inside [ObjectExtension]. [map] is the current
  /// nesting element in the object structure. [level] defines the current
  /// nesting level in the object structure.
  ///
  /// The [stringQuote], [keyQuote], [withTypes] and [tabWidth] parameters are
  /// described in the reference for the [toStructuredString()] method.
  ///
  String _mapToStructuredString(
      {required Map map,
      required String stringQuote,
      required String keyQuote,
      required bool withTypes,
      required tabWidth,
      required int level}) {
    final tab = ' ' * tabWidth;
    var str = '{\n';

    for (final entry in map.entries) {
      final value = entry.value;

      if (entry.key is String) {
        str += tab * (level + 1) + keyQuote + entry.key + keyQuote + ': ';
      } else {
        str += tab * (level + 1) + entry.key.toString() + ': ';
      }
      str += _toStructuredString(
          value: value,
          parentIsSimpleList: false,
          stringQuote: stringQuote,
          keyQuote: keyQuote,
          withTypes: withTypes,
          tabWidth: tabWidth,
          level: level + 1);
      if (entry.key == map.entries.last.key) {
        str += '\n';
      } else {
        str += ',\n';
      }
    }

    return str + tab * level + '}';
  }

  ///
  /// Converts [List] type object to a human-readable structured string.
  ///
  /// Closed method, used only inside [ObjectExtension]. [list] is the current
  /// nesting element in the object structure. [level] defines the current
  /// nesting level in the object structure.
  ///
  /// The [stringQuote], [keyQuote], [withTypes] and [tabWidth] parameters are
  /// described in the reference for the [toStructuredString()] method.
  ///
  String _listToStructuredString(
      {required List list,
      required String stringQuote,
      required String keyQuote,
      required bool withTypes,
      required tabWidth,
      required int level}) {
    final tab = ' ' * tabWidth;
    final simple = list.runtimeType.toString() == 'List<Object>' ? false : true;
    var str = simple ? '[ ' : '[\n';

    for (final value in list) {
      if (!simple) str += tab * (level + 1);
      str += _toStructuredString(
          value: value,
          parentIsSimpleList: simple,
          stringQuote: stringQuote,
          keyQuote: keyQuote,
          withTypes: withTypes,
          tabWidth: tabWidth,
          level: level + 1);
      str += _divider(value == list.last, simple);
    }

    if (simple) return str + ']';
    return str + tab * level + ']';
  }

  ///
  /// Converts [Set] type object to a human-readable structured string.
  ///
  /// Closed method, used only inside [ObjectExtension]. [set] is the current
  /// nesting element in the object structure. [level] defines the current
  /// nesting level in the object structure.
  ///
  /// The [stringQuote], [keyQuote], [withTypes] and [tabWidth] parameters are
  /// described in the reference for the [toStructuredString()] method.
  ///
  String _setToStructuredString(
      {required Set set,
      required String stringQuote,
      required String keyQuote,
      required bool withTypes,
      required tabWidth,
      required int level}) {
    final tab = ' ' * tabWidth;
    final simple =
        set.toList().runtimeType.toString() == 'List<Object>' ? false : true;
    var str = simple ? '{ ' : '{\n';

    for (final value in set) {
      if (!simple) str += tab * (level + 1);
      str += _toStructuredString(
          value: value,
          parentIsSimpleList: simple,
          stringQuote: stringQuote,
          keyQuote: keyQuote,
          withTypes: withTypes,
          tabWidth: tabWidth,
          level: level + 1);
      str += _divider(value == set.last, simple);
    }

    if (simple) return str + '}';
    return str + tab * level + '}';
  }

  ///
  /// Forms a divider between entries. The type of divider depends on its
  /// location, as well as on the complexity of the parent object.
  ///
  String _divider(bool last, bool simple) {
    if (last) {
      if (simple) return ' ';
      return '\n';
    }
    if (simple) return ', ';
    return ',\n';
  }
}
